# klive: kernel live user testing feedback
# Copyright (C) 2005-2006  Andrea Arcangeli <andrea@cpushare.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# You should start the daemon at boot with: 'twistd -oy klive.tac'
# The info sent in cleartext over the network is "os.uname()[2:5]"
# plus parts of: /proc/meminfo, /proc/cpuinfo, /sys/bus, /proc/stat,
# /proc/mounts, lspci -n, /proc/modules, /proc/config.gz.

# You can tweak the CONFIG_* variables before running this script.

# When set to True it will allow you to identify your systems reliably
# by creating a file called klive.id- in the same directory where this
# program is located, you can copy it to your other systems if you
# want to track all your systems under the same ID. If you lose the
# klive.id- file you will not be able to identify your systems anymore.
# If unsure leave it disabled (default). It can also be forced disabled
# or enabled by setting the environemnt variable 'KLIVE_SYSTEM_ID'
# respectively to 0 or 1.
try:
	import os
	CONFIG_SYSTEM_ID = os.environ['KLIVE_SYSTEM_ID']
	if CONFIG_SYSTEM_ID == '0':
		CONFIG_SYSTEM_ID = False
	elif CONFIG_SYSTEM_ID == '1':
		CONFIG_SYSTEM_ID = True
	else:
		raise
except:
	CONFIG_SYSTEM_ID = False
	#CONFIG_SYSTEM_ID = True

# When not None, it selects the directory where to keep the klive.id-*
# file. It can also be overridden with the environemnt variable
# KLIVE_SYSTEM_ID_DIR.
try:
	import os
	CONFIG_SYSTEM_ID_DIR = os.environ['KLIVE_SYSTEM_ID_DIR']
except KeyError:
	CONFIG_SYSTEM_ID_DIR = None
	#CONFIG_SYSTEM_ID_DIR = '/somepath/'

# When set to True (default) it will send the filesystem information
# in /proc/mounts to the server.
CONFIG_FS = True

# When set to True (default) it will send the `lspci -n` information
# to the server.
CONFIG_PCI_ID = True

# When set to True (default) it will send the /proc/modules information
# to the server.
CONFIG_MODULES = True


PUSH_INTERVAL = 60*10 # start with one push every 10 minutes
PUSH_INTERVAL_MAX = 60*60*24 # max out the backoff at 1 push per day
PUSH_INTERVAL_BACKOFF = 1.25
SERVER = 'klive.cpushare.com'

if 0:
	PUSH_INTERVAL = 1
	PUSH_INTERVAL_MAX = 10
	SERVER = 'localhost'

PORT = 4921

import sys, os, struct, md5, re, glob, stat, zlib, socket

from twisted.internet.protocol import DatagramProtocol
from twisted.internet import reactor

PROTOCOL_VERSION = 1

# Object is to track Linux Testing
assert os.uname()[0] == 'Linux'

def get_mhz():
	alpha = re.search(r' (\d+)',
			  os.popen(r"grep 'cycle frequency \[Hz\]' /proc/cpuinfo | head -n 1").read())
	if alpha:
		return int(alpha.group(1))/1000000

	sparc64 = re.search(r': (\w{16})$',
			  os.popen("grep -i ClkTck /proc/cpuinfo | head -n 1").read())
	if sparc64:
		return int(sparc64.group(1), 16)/1000000

	all = re.search(r' (\d+)\.?\d?',
			os.popen("grep -i mhz /proc/cpuinfo | grep -v 'model name' | head -n 1").read())
	if all:
		return int(all.group(1))

	not_sun4m = os.system('grep -q ": sun4m" /proc/cpuinfo')
	if not not_sun4m:
		sun4m = re.search(r' (\d+)\.?\d?',
				  os.popen("grep -i BogoMips /proc/cpuinfo | head -n 1").read())
		if sun4m:
			return int(sun4m.group(1))

	raise 'Your /proc/cpuinfo is not compatible, please send it in a bugreport'

def str_append(s):
	return struct.pack('!H', len(s)) + s

def gen_random_hash():
	return md5.new(file('/dev/urandom').read(100) + os.uname()[1]).digest()	

def hexdigest(digest):
	import string
	hexdigest = ''
	for i in digest:
		i = ord(i)
		hexdigest += string.hexdigits[(i >> 4) & 0xf]
		hexdigest += string.hexdigits[i & 0xf]
	return hexdigest

KERNEL = ''.join([(chr(len(x)) + x) for x in os.uname()[2:5]])
if PROTOCOL_VERSION > 0:
	proc_version = file('/proc/version').read()
	KERNEL += chr(len(proc_version)) + proc_version

	if CONFIG_FS:
		fs = {}
		for _fs in [ re.search(r'^[^\s]+\s[^\s]+\s([^\s]+)', x).group(1)
			     for x in file('/proc/mounts').readlines() ]:
			if _fs not in ['rootfs', 'proc', 'devpts']:
				fs[_fs] = None
		fs = ' '.join(fs.keys())
		KERNEL += str_append(fs)
	else:
		KERNEL += str_append('')

	if CONFIG_PCI_ID:
		os.environ['PATH'] += ':/sbin:/usr/sbin'
		pci_id = os.popen('lspci -n').readlines()
		pci_id = ''.join([ re.sub('.* Class ', '', x) for x in pci_id ])
		KERNEL += str_append(pci_id)
	else:
		KERNEL += str_append('')

	if CONFIG_MODULES:
		try:
			modules = []
			modules = ' '.join([ re.search(r'^([^\s]+)', x).group(1)
					     for x in file('/proc/modules').readlines() ])
		except IOError:
			modules = ''
		KERNEL += str_append(modules)
	else:
		KERNEL += str_append('')

	if CONFIG_SYSTEM_ID:
		prefix = 'klive.id-'
		if CONFIG_SYSTEM_ID_DIR is not None:
			prefix = os.path.join(CONFIG_SYSTEM_ID_DIR, prefix)
		system_id = glob.glob(prefix + '*')
		if len(system_id) > 1:
			print >>sys.stderr, 'ERROR: more than one klive.id installed'
			sys.exit(1)
		elif len(system_id) == 1:
			system_id = system_id[0]
			if os.stat(system_id)[stat.ST_UID] != os.getuid():
				print >>sys.stderr, 'ERROR: klive.id has been installed by a different user'
				sys.exit(1)
			digest = file(system_id).read()
			if hexdigest(digest) != re.search(prefix + '(.*)$', system_id).group(1):
				print >>sys.stderr, 'ERROR: klive.id has been corrupted'
				sys.exit(1)
		else:
			digest = gen_random_hash()
			system_id = prefix + hexdigest(digest)
			handle = file(system_id, 'w')
			handle.write(digest)
			del handle
		KERNEL += chr(len(digest)) + digest
	else:
		KERNEL += chr(0)

SESSION = gen_random_hash()
RAM = int(re.match(r'MemTotal:\s+(\d+) kB',
		   os.popen('grep MemTotal /proc/meminfo').read()).group(1)) / 1024
CPUS = int(os.popen("egrep '^cpu[0-9]+' /proc/stat | wc -l").read())
MHZ = get_mhz()
DEVICES = int(os.popen('find /sys/bus/*/devices 2>/dev/null | wc -l').read())

class klive_protocol(DatagramProtocol):
	connected = False
	timer = None

	def __init__(self, seconds, push_interval):
		self.seconds = seconds
		self.push_interval = push_interval

	def backoff(self):
		self.seconds += self.push_interval
		self.push_interval *= PUSH_INTERVAL_BACKOFF
		self.push_interval = min(self.push_interval, PUSH_INTERVAL_MAX)

	def push_ip_end(self):
		self.timer = reactor.callLater(self.push_interval, self.push)
		self.backoff()

	def push_ip_success(self, ip):
		def restart_proto():
			reactor.listenUDP(0, klive_protocol(self.seconds, self.push_interval))
		def restart_proto_delay(x):
			reactor.callLater(self.push_interval, restart_proto)
			self.backoff()

		if not self.connected:
			try:
				self.transport.connect(ip, PORT)
			except socket.error, se:
				print 'transport.connect', se
				self.transport.stopListening().addCallback(restart_proto_delay)
				return
			else:
				self.connected = True

		data = struct.pack('!I', PROTOCOL_VERSION)

		payload = struct.pack('!IIIII', self.seconds, RAM, CPUS, MHZ, DEVICES)
		payload += KERNEL + SESSION

		if not PROTOCOL_VERSION:
			data += payload
		elif PROTOCOL_VERSION > 0:
			data += str_append(zlib.compress(payload, 9))

		try:
			self.transport.write(data)
			#print 'Packet sent successfully'
		except socket.error, se:
			print 'transport.write', se
			self.transport.stopListening().addCallback(restart_proto_delay)
			return

		self.push_ip_end()

	def push_ip_failed(self, failure):
		print failure.getErrorMessage()
		self.push_ip_end()

	def push(self):
		self.timer = None
		reactor.resolve(SERVER).addCallbacks(self.push_ip_success, self.push_ip_failed)

	def startProtocol(self):
		self.push()

	def stopProtocol(self):
		if self.timer:
			self.timer.cancel()

	def connectionRefused(self):
		print 'klive server refused the packet'

from twisted.internet import app
application = app.Application('klive_client')
application.listenUDP(0, klive_protocol(0, PUSH_INTERVAL))
